# python3
# coding=<UTF-8>

import re


class Target:
    def __init__(self, text, idxs, meta, tags, transl=None, lang=None):
        """
        text: str: full sentence / document
        idxs: tuple (l, r): target idxs in self.text -> self.text[l:r]
        meta: str: sentence / document info (if exists)
        tags: dict?: target tags
        transl: str: text translation (only for parallel corpora)
        lang: str: translation language (only for parallel corpora)
        """
        self.text = text
        self.idxs = idxs
        self.meta = meta
        self.tags = tags
        self.transl = transl
        self.lang = lang
        
    def __str__(self):
        return 'Target(%s, %s)' \
                % (self.text[self.idxs[0]: self.idxs[1]],
                   self.meta
                   )

    __repr__ = __str__
        
    def __get_kwic_wlvl_idx(self):
        """
        get word level target index
        """
        _t = '<TARGET>%s<TARGET>' \
              % self.text[self.idxs[0]: self.idxs[1]]
        _s = '%s%s%s' % (self.text[: self.idxs[0]],
                         _t,
                         self.text[self.idxs[1]:]
                        )
        for i, tkn in enumerate(_s.split()):
            if _t in tkn:
                return i
        raise ValueError('no target found, check `level`')
    
    def __fix_kwic(self, l, c, r):
        """
        handle punctuation outside the target
        """
        if re.search('[\W]', c) is not None:
            l_punct = re.search('(^[\W]*)', c).group(1)
            r_punct = re.search('([\W]*)$', c).group(1)
            c = re.sub('[\W]', '', c)
            l += l_punct
            r = r_punct + r
        return (l, c, r)
    
    def kwic(self, left, right, level='word'):
        """
        level: ['word', 'char']: if "word" - split by words
                                 if "char" - split by characters
        """
        if level not in ['word', 'char']:
            raise ValueError('got invalid `level` "%s"' % level)
        
        if ' ' not in self.text:
            level = 'char'
        
        if level == 'word':
            tkns = self.text.split()
            idx = self.__get_kwic_wlvl_idx()
            if idx - left < 1:
                return self.__fix_kwic(' '.join(tkns[: idx]),
                                       tkns[idx],
                                       ' '.join(tkns[idx+1: idx+right+1])
                                      )
            return self.__fix_kwic(' '.join(tkns[idx-left: idx]),
                                   tkns[idx],
                                   ' '.join(tkns[idx+1: idx+right+1])
                                  )

        else:
            if self.idxs[0] - left < 1:
                return (self.text[: self.idxs[0]],
                        self.text[self.idxs[0]: self.idxs[1]],
                        self.text[self.idxs[1]:self.idxs[1]+right]
                       )
            return (self.text[self.idxs[0]-left: self.idxs[0]],
                    self.text[self.idxs[0]: self.idxs[1]],
                    self.text[self.idxs[1]: self.idxs[1]+right]
                   )



# python3
# coding=<UTF-8>

from collections import Iterable
import warnings
from time import sleep
from tqdm import tqdm
from .result import Result
from .corpora import *


functions = {
             'rus': rus_corpus,
             'bam': bam_corpus,
             'emk': emk_corpus,
             'zho': zho_corpus,
             'rus_parallel': rus_parallel_corpus
            }


class Corpus:
    def __init__(self, language, sleep_time=1, sleep_each=5):
        """
        language: str: language alias
        sleep_time: int: sleeping time in seconds
        sleep_each: int: sleep after each `sleep_each` request
        """
        
        self.language = language
        self.__corpus = functions[self.language] 
        self.doc = self.__corpus.__doc__
        
        self.results = list()
        self.failed = list()
        self.__retry_flag = False
        
        self.__warn = 'Nothing found for query "%s".\n' \
                      'Call `retry_failed` method to retry failed queries'
        self.__pbar_desc = '"%s"'
        self.__type_except = 'Argument `query` must be of type <str> or iterable, got <%s>'

        if sleep_each < 1:
            raise ValueError('Argument `sleep_each` must  be >= 1')
            
        self.sleep_each = sleep_each
        self.sleep_time = sleep_time

    def search(self, query, *args, **kwargs):
        """
        query: str: query
        for arguments see `params_container.Container`
        """
        
        if isinstance(query, str):
            query = [query]
        
        if not isinstance(query, Iterable):
            raise TypeError(self.__type_except % type(query))
            
        if args:
            progress_total = args[0]
        elif 'numResults' in kwargs:
            progress_total = kwargs['numResults']
        else:
            progress_total = 100
        
        _results = list()
        
        for q in query:
            self.parser = self.__corpus.PageParser(q, *args, **kwargs)
            _r = Result(self.language, self.parser.__dict__)
            q_desc = self.__pbar_desc % q
                
            for t in tqdm(self.parser.extract(),
                          total=progress_total,
                          unit='docs',
                          desc=q_desc):
                _r.add(t)
                if _r.N % self.sleep_each == 0:
                    sleep(self.sleep_time)
            
            _results.append(_r)
            if _r.N < 1:
                warnings.warn(self.__warn % q)
                if not self.__retry_flag:
                    self.failed.append(_r)
        
        if not self.__retry_flag:
            self.results.extend(_results)
        
        return _results

    def retry_failed(self):
        """
        Calls `.search()` for failed queries stored in `.failed`
        
        ISSUE:
        if `_r` got successfully retrieved here,
        its empty `Result` is still left in `Corpus.results` 
        """
        if self.failed:
            self.__retry_flag = True
            _pos = list()
            _neg = list()
            
            for r in self.failed:
                _r = self.search(r.query, **r.params)[0]
                if _r.N > 0:
                    _pos.append(_r)
                else:
                    _neg.append(_r)
            
            self.failed = _neg[:]
            self.results.extend(_pos)
            self.__retry_flag = False
            
            return _pos
        
        else:
            return []



# python3
# coding=<UTF-8>

import re
import csv


class Result:
    def __init__(self, language, query_params):
        """
        language: str: language
        query_params: dict: __dict__ of used parser
        """
        self.results = list()
        self.N = 0
        self.lang = language
        self.query = query_params['query']
        self.params = {k: query_params[k] \
                      for k in query_params \
                      if k[0] != '_' and k not in ['page', 'query']
                     }
                      
        self.__header = ('index', 'text')
        self.__kwic_header = ('index', 'left', 'center', 'right')
        self.__not_allowed = '/\\?%*:|"<>'
    
    def add(self, x):
        self.results.append(x)
        self.N += 1
    
    def __str__(self):
        return 'Result(query=%s, N=%s, params=%s)' \
                % (self.query,
                   self.N,
                   self.params
                  )
    
    __repr__ = __str__
    
    def __iter__(self):
        return iter(self.results)
        
    def __getitem__(self,key):
        return self.results[key]
        
    def __setitem__(self,key,val):
        self.results[key] = val

    def __delitem__(self, key):
        del self.results[key]
    
    def export_csv(self, filename=None, header=True, sep=';'):
        if filename is None:
            filename = '%s_%s_results.csv' \
                        % (self.lang,
                           re.sub(self.__not_allowed, '', self.query)
                          )
        
        with open(filename,'w',encoding='utf-8-sig') as f:
            writer = csv.writer(f, delimiter=sep, quotechar='"',
                                quoting=csv.QUOTE_MINIMAL, lineterminator='\n'
                               )
            
            if self.params['kwic']:
                if header:
                    writer.writerow(self.__kwic_header)
                
                nLeft = self.params['nLeft'] 
                nRight = self.params['nRight']
                
                if nLeft is None:
                    nLeft = 10
                    
                if nRight is None:
                    nRight = 10
                    
                for i, t in enumerate(self.results):
                    writer.writerow((i + 1, *t.kwic(nLeft, nRight)))
            
            else:
                if header:
                    writer.writerow(self.__header)
                
                for i, t in enumerate(self.results):
                    writer.writerow((i + 1, t.text))
                    
    def clear(self):
        self.results = list()



# python3
# coding=<UTF-8>


class Container:
    def __init__(self, query, numResults=100, kwic=True, nLeft=None, nRight=None,
                 subcorpus=None, tag=False, queryLanguage=None,
                 start=0, writingSystem=None):
        self.query = query
        self.numResults = numResults
        self.kwic = kwic
        self.nLeft = nLeft
        self.nRight = nRight
        self.queryLanguage = queryLanguage
        self.subcorpus = subcorpus
        self.tag = tag
        self.start = start
        self.writingSystem = writingSystem



from setuptools import setup

setup(name='lingcorpora',
      version='1.0',
      description='API for corpora',
      url='https://github.com/lingcorpora/lingcorpora.py',
      author='Alexey Koshevoy, Artem Kopetsky, Ekaterina Gerasimenko, Maria Terekhina',
      license='MIT',
      author_email='alexeykochevoy@gmail.com',
      packages=['lingcorpora'],
      zip_safe=False,
      keywords = ['corpora', 'api', 'language'],
      install_requires=['bs4', 'requests', 'lxml', 'tqdm'])
